#ifndef EQUIS_IMAGE_IMAGE_HPP
#define EQUIS_IMAGE_IMAGE_HPP

#include "pixel.hpp"

EQUIS_ENTER_NS2(equis,image);

template < typename T >
class PackedImage {
public:
	typedef T Pixel;

	PackedImage(void* data, long rows, long cols)
		: m_data(reinterpret_cast<Pixel*>(data)), m_rows(rows), m_cols(cols)
	{}

	long rows() const { return m_rows; }
	long cols() const { return m_cols; }

	Pixel& operator () (long row, long col) {
		EQUIS_DEBUG_ASSERT(row >= 0 && row < m_rows);
		EQUIS_DEBUG_ASSERT(col >= 0 && col < m_cols);
		return m_data[row * m_cols + col];
	};

private:
	T* m_data;
	long m_rows;
	long m_cols;
};

template < typename T >
class ConstPackedImage {
public:
	typedef T Pixel;

	ConstPackedImage(const void* data, long rows, long cols)
		: m_data(reinterpret_cast<const Pixel*>(data)), m_rows(rows), m_cols(cols)
	{}

	long rows() const { return m_rows; }
	long cols() const { return m_cols; }

	Pixel operator () (long row, long col) const {
		EQUIS_DEBUG_ASSERT(row >= 0 && row < m_rows);
		EQUIS_DEBUG_ASSERT(col >= 0 && col < m_cols);
		return m_data[row * m_cols + col];
	};

private:
	const T* m_data;
	long m_rows;
	long m_cols;
};

class ConstI420Image {
public:
	typedef PixelY8U8V8 Pixel;

	ConstI420Image(const void* data, long rows, long cols)
		: m_rows(rows), m_cols(cols), m_half_rows((rows+1)/2), m_half_cols((cols+1)/2)
	{
		m_ys = reinterpret_cast<const uint8*>(data);
		m_us = m_ys + m_rows * m_cols;
		m_vs = m_us + m_half_rows * m_half_cols;
	}

	long rows() const { return m_rows; }
	long cols() const { return m_cols; }

	Pixel operator () (long row, long col) const {
		Pixel dst;
		dst.y = m_ys[row * m_cols + col];
		dst.u = m_us[(row/2) * m_half_cols + (col/2)];
		dst.v = m_vs[(row/2) * m_half_cols + (col/2)];
		return dst;
	};

private:
	const uint8* m_ys;
	const uint8* m_us;
	const uint8* m_vs;
	long m_rows;
	long m_cols;
	long m_half_rows;
	long m_half_cols;
};

struct A8R8G8B8 {
	typedef PixelA8R8G8B8 Pixel;
	typedef PackedImage<Pixel> Image;
	typedef ConstPackedImage<Pixel> ConstImage;
};

struct R8G8B8 {
	typedef PixelR8G8B8 Pixel;
	typedef PackedImage<Pixel> Image;
	typedef ConstPackedImage<Pixel> ConstImage;
};

struct Y8 {
	typedef PixelY8 Pixel;
	typedef PackedImage<Pixel> Image;
	typedef ConstPackedImage<Pixel> ConstImage;
};

struct Y16 {
	typedef PixelY16 Pixel;
	typedef PackedImage<Pixel> Image;
	typedef ConstPackedImage<Pixel> ConstImage;
};

struct I420 {
	typedef PixelY8U8V8 Pixel;
	typedef ConstI420Image ConstImage;
};

template < typename DstTraits, typename SrcTraits > inline
void convertImage(
	void* dst, const void* src,
	long rows, long cols, bool vflip);

EQUIS_LEAVE_NS2(equis,image);

#define IN_EQUIS_IMAGE_IMAGE_HPP
#include "image.inl"
#undef IN_EQUIS_IMAGE_IMAGE_HPP

#endif /* EQUIS_IMAGE_IMAGE_HPP */
