#ifndef EQUIS_MATH_HPP
#define EQUIS_MATH_HPP

EQUIS_ENTER_NS1(equis);

/// @name Pointer manipulation functions
//@{

/// @brief Offsets a pointer by the specified number of bytes.
template < typename T > inline T*
byte_offset(T* base, ptrdiff_t offset);

/// @brief Offsets a constant pointer by the specified number of bytes.
template < typename T > inline const T*
byte_offset(const T* base, ptrdiff_t offset);

//@}

/// @brief Determines the number of set bits.
/// The type must be an unsigned integer.
template < typename T > inline int
count_one_bits(T value);

/// @brief Determines the number of clear bits.
/// The type must be an unsigned integer.
template < typename T > inline int
count_zero_bits(T value);

/// @brief Determines the number of high clear bits before the first set bit.
/// The type must be an unsigned integer.
template < typename T > inline int
leading_zero_bits(T value);

/// @brief Determines the number of high set bits until the first clear bit.
/// The type must be an unsigned integer.
template < typename T > inline int
leading_one_bits(T value);

/// @brief Determines the number of low clear bits until the first set bit.
/// The type must be an unsigned integer.
template < typename T > inline int
trailing_zero_bits(T value);

/// @brief Determines the number of low set bits until the first clear bit.
/// The type must be an unsigned integer.
template < typename T > inline int
trailing_one_bits(T value);

/// @brief Generates the smallest power of two greater than or equal to the number.
/// The type must be an unsigned integer.
template < typename T > inline T
power_of_two(T value);

/// @brief Generates the smallest power of two greater than the number.
/// Returns zero if no such value is representable.
/// The type must be an unsigned integer.
template < typename T > inline T 
next_power_of_two(T value);

/// @brief Multiplies two values and returns the result without overflow.
template < typename T > inline typename numeric_traits<T>::promoted_type
mul(T a, T b);

/// @brief Multiplies two values and returns the high half of the result.
/// The type must be an unsigned integer.
template < typename T > inline T
mul_hi(T a, T b);

/// Align a pointer or byte count to the specified mask.
template < typename T > inline T
align(T value, T mask);

/// @brief Clamps a value to the specified range.
template < typename T > inline T
clamp(T value, T minimum, T maximum);

/// @brief Clamps a value to the specified range.
template < typename T, typename Less > inline T
clamp(T value, T minimum, T maximum, Less func = Less());

EQUIS_LEAVE_NS1(equis);

#define IN_EQUIS_MATH_HPP
#include "math.inl"
#undef IN_EQUIS_MATH_HPP

#endif /* EQUIS_MATH_HPP */