#ifndef EQUIS_WIN32_COMPTR_HPP
#define EQUIS_WIN32_COMPTR_HPP

/// A safe COM pointer.
template < typename T >
class ComPtr {
public:
    /// Encapsulated type.
    typedef T Type;

    /// Gets the UUID of the encapsulated type, if defined.
    static REFIID GetUuid() { return __uuidof(T); }

    /// Creates a NULL pointer.
    ComPtr();

    /// Attaches to the specified pointer.
    ComPtr(const ComPtr<T>& src);

    /// Attaches to the specified pointer.
    ComPtr(T* ptr);

    /// Detaches this pointer.
    ~ComPtr();

    /// Determines if this pointer is NULL.
    bool IsNull() const;

    /// Determines if this pointer is not NULL.
    bool IsNotNull() const;

    /// Invokes an operation on this pointer. In debug mode, invoking an
    /// operation on a NULL pointer will trigger a breakpoint.
    T* operator -> () const;

    /// Gets the underlying pointer without adding a reference.
    operator T* () const;

    /// Gets a new reference of the underlying pointer.
    T* Get() const;

    /// Gets the underlying pointer without adding a reference.
    T* GetWithoutAddRef() const;

    /// Attaches to the specified pointer.
    ComPtr<T>& operator = (const ComPtr<T>& src);

    /// Attaches to the specified pointer.
    ComPtr<T>& operator = (T* ptr);

    /// Attaches to the specified pointer.
    void Attach(T* ptr);

    /// Attaches to the specified pointer without adding a reference.
    void AttachWithoutAddRef(T* ptr);

    /// Attaches to the specified pointer without releasing it later.
    void AttachWithoutRelease(T* ptr);

    /// Attaches to the specified pointer without adding a reference or
    /// releasing it later.
    void AttachWithoutAddRefOrRelease(T* ptr);

    /// Detaches this pointer, and sets it to NULL.
    void Detach();

    /// Detaches this pointer without releasing it, and sets it to NULL.
    T* DetachWithoutRelease();

    /// Detaches this pointer, sets it to NULL, and returns its address.
    T** DetachAndGetPointer();

    /// Detaches this pointer, sets it to NULL, and returns its address.
    void** DetachAndGetVoidPointer();

private:
    T* m_ptr;
    bool m_noRelease;
};

#define IN_EQUIS_WIN32_COMPTR_HPP
#include "ComPtr.inl"
#undef IN_EQUIS_WIN32_COMPTR_HPP

#endif /* EQUIS_WIN32_COMPTR_HPP */