
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Content;
using Equis.Xna.Input;
using Microsoft.Xna.Framework.Audio;
using System.IO;
using System.Diagnostics;
using System.Threading;

namespace Demos.Rocks {
    public class RocksGame : Game {
        public static readonly String[] AsteroidModels = new String[] {
            @"Models\asteroid1", @"Models\asteroid2"
        };

        public static readonly String AudioEngineFile = @"Audio\rocks.xgs";
        public static readonly String SoundBankFile = @"Audio\rocks.xsb";
        public static readonly String WaveBankFile = @"Audio\rocks.xwb";

        public static readonly String SaucerModel = @"Models\p1_saucer";
        public static readonly String LargeAsteroidModel = @"Models\asteroid1";
        public static readonly String SmallAsteroidModel = @"Models\asteroid2";
        public static readonly String RocketModel = @"Models\p1_rocket";

        public static readonly String LargeAsteroidTexture = @"Textures\asteroid1";
        public static readonly String SmallAsteroidTexture = @"Textures\asteroid2";

        private readonly GraphicsDeviceManager graphicsManager;
        private readonly Ftir ftir;
        private readonly ContentManager contentManager;
        private readonly AudioEngine audioEngine;
        private readonly WaveBank waveBank;
        private readonly SoundBank soundBank;

        private String baseTitle;

        private KeyboardState oldKeyboardState;
        private MouseState oldMouseState;

        private Matrix projection;
        private Matrix view;

        private readonly List<Shape> shapes = new List<Shape>();
        private Particles particles;
        private Particle scoreParticle;
        private Shape saucer;
        private Background background;
        private readonly Dictionary<int, GuidedShape> missiles
            = new Dictionary<int, GuidedShape>();
        private int score = 0;

        private bool showEvents = false;
        private bool showInput = false;
        private bool showOutput = false;
        private Texture2D previewTexture;

        private Shape calibrationAsteroid;
        private int calibrationIndex = -2;
        private double calibrationLastEvent = 0.0;
        private Vector2[] calibrationExpected;
        private Vector2[] calibrationActual;

        private TimeSpan lastEvent = TimeSpan.Zero;
        private TimeSpan nextAsteroidSpawn = TimeSpan.Zero;
        private int viewportWidth;
        private bool selfPlay;
        private int selfPlayId = 1000000;

        public RocksGame(bool muteSound) {
            Window.Title = "FTIR Demo";

            graphicsManager = new GraphicsDeviceManager(this);
            graphicsManager.IsFullScreen = true;

            ftir = Ftir.FromCapture(null);

            contentManager = new ContentManager(Services, "Content");

            if (muteSound) {
                audioEngine = null;
                waveBank = null;
                soundBank = null;
            } else {
                audioEngine = new AudioEngine(Path.Combine("Content", AudioEngineFile));
                waveBank = new WaveBank(audioEngine, Path.Combine("Content", WaveBankFile));
                soundBank = new SoundBank(audioEngine, Path.Combine("Content", SoundBankFile));

            }

            shapes = new List<Shape>();
        }

        private bool Calibrating {
            get {
                return calibrationIndex >= 0 &&
                    calibrationIndex < calibrationExpected.Length;
            }
        }

        public GraphicsDeviceManager GraphicsManager {
            get {
                return graphicsManager;
            }
        }

        public ContentManager ContentManager {
            get {
                return contentManager;
            }
        }

        public Matrix Projection {
            get {
                return projection;
            }
        }

        public Matrix View {
            get {
                return view;
            }
        }

        public IEnumerable<Shape> Shapes {
            get {
                return shapes;
            }
        }

        private void OnPreparingDeviceSettings(Object sender, PreparingDeviceSettingsEventArgs args) {
            PresentationParameters present = args.GraphicsDeviceInformation.PresentationParameters;
            present.PresentationInterval = PresentInterval.One;
            present.MultiSampleType = MultiSampleType.None;
            present.SwapEffect = SwapEffect.Discard;

            if (present.IsFullScreen) {
                present.BackBufferWidth = args.GraphicsDeviceInformation.Adapter.CurrentDisplayMode.Width;
                present.BackBufferHeight = args.GraphicsDeviceInformation.Adapter.CurrentDisplayMode.Height;
            }
        }

        protected override void Initialize() {
            Window.AllowUserResizing = true;
            graphicsManager.PreferMultiSampling = false;
            if (ftir != null) {
                while (ftir.Format.Width == 0) {
                    Thread.Sleep(500);
                }
                graphicsManager.PreferredBackBufferWidth = ftir.Format.Width;
                graphicsManager.PreferredBackBufferHeight = ftir.Format.Height;
            }
            graphicsManager.PreparingDeviceSettings += OnPreparingDeviceSettings;
            graphicsManager.SynchronizeWithVerticalRetrace = true;
            graphicsManager.ApplyChanges();

            baseTitle = Window.Title;
            IsMouseVisible = true;

            saucer = new Shape(this);
            saucer.ModelPath = SaucerModel;
            saucer.Position = Vector3.Down * 7000.0f;
            saucer.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Left, MathHelper.ToRadians(90.0f));
            saucer.AmbientLightColor = Vector3.One * 0.2f;
            saucer.DiffuseLightColor = Vector3.One * 0.4f;
            saucer.SpecularLightColor = Vector3.One * 0.4f;
            shapes.Add(saucer);

            particles = new Particles(this);
            scoreParticle = particles.AddText(score.ToString(), new Vector2(0, 0));
            scoreParticle.Origin = new Vector2(1, 1);
            scoreParticle.StartScale = scoreParticle.EndScale = 0.75f;
            shapes.Add(particles);

            calibrationAsteroid = new Shape(this);
            calibrationAsteroid.ModelPath = SmallAsteroidModel;
            calibrationAsteroid.TexturePath = SmallAsteroidTexture;
            calibrationAsteroid.Scale = new Vector3(0.5f, 0.5f, 0.5f);
            calibrationAsteroid.AngularVelocity = Quaternion.CreateFromYawPitchRoll(
                MathHelper.ToRadians(45.0f),
                MathHelper.ToRadians(30.0f),
                MathHelper.ToRadians(60.0f));

            background = new Background(this);
            background.TexturePath = @"Textures\B1_stars";
            if (ftir != null) {
                //background.Video = new Video[] { ftir.InputPreview, ftir.ChangePreview };
            }
            shapes.Insert(0, background);

            base.Initialize();
        }

        protected override void LoadContent() {
            foreach (Shape shape in shapes) {
                shape.Setup();
            }

            if (ftir != null) {
                ftir.Start();
            }
        }

        protected override void UnloadContent() {
            if (ftir != null) {
                ftir.Stop();
            }

            contentManager.Unload();
        }

        private void UpdateCalibrationAsteroid() {
            if (Calibrating) {
                calibrationAsteroid.Position = GameTools.UnprojectToZeroPlane(
                    this, calibrationExpected[calibrationIndex]);
                if (soundBank != null) {
                    soundBank.PlayCue("calibrate_beep");
                }
            } else {
                if (ftir != null && calibrationExpected != null && calibrationIndex == calibrationExpected.Length) {
                    int width = GraphicsDevice.Viewport.Width;
                    int height = GraphicsDevice.Viewport.Height;
                    for (int i = 0; i < calibrationExpected.Length; ++i) {
                        calibrationExpected[i].X /= width;
                        calibrationExpected[i].Y /= height;
                    }
                    ftir.Calibrate(calibrationExpected, calibrationActual);
                }

                calibrationAsteroid.Deleted = true;
                if (soundBank != null) {
                    soundBank.PlayCue("calibrate_done");
                }

                calibrationIndex = -2;
                calibrationExpected = null;
                calibrationActual = null;
            }
        }

        protected override void Update(GameTime gameTime) {
            GraphicsDevice graphics = graphicsManager.GraphicsDevice;
            viewportWidth = graphics.Viewport.Width;

            if (ftir != null) {
                Window.Title = String.Format("{0} (FPS: render={1:F} format={2:F} camera={3:F})",
                    baseTitle,
                    graphics.DisplayMode.RefreshRate,
                    ftir.Format.FramesPerSecond,
                    ftir.RealFramesPerSecond);
            }
            // Initialize parameters

            float aspectRatio = (float)graphics.Viewport.Width / (float)graphics.Viewport.Height;
            scoreParticle.Position = new Vector2(
                graphics.Viewport.Width - 20,
                graphics.Viewport.Height - 20);

            if (showEvents || Calibrating) {
                scoreParticle.Text = "";
            } else if (score <= 0) {
                scoreParticle.Text = "0";
            } else {
                scoreParticle.Text = score.ToString();
            }


            view = Matrix.CreateLookAt(
                Vector3.Backward * 10000.0f,
                Vector3.Zero, Vector3.Up);

            projection = Matrix.CreateOrthographic(20000.0f, 20000.0f / aspectRatio, 1.0f, 50000.0f);

            Vector3 saucerPosition = saucer.Position;
            saucerPosition.Y = -8000 / aspectRatio;
            saucer.Position = saucerPosition;

            if (calibrationIndex == -1) {
                // Starting calibration.

                foreach (Shape shape in shapes) {
                    if (Array.IndexOf(AsteroidModels, shape.ModelPath) >= 0) {
                        shape.Deleted = true;
                    }
                }

                calibrationAsteroid.Deleted = false;
                calibrationAsteroid.Setup();
                if (!shapes.Contains(calibrationAsteroid)) {
                    shapes.Add(calibrationAsteroid);
                }

                int offsetX = graphics.Viewport.Width / 4;
                int offsetY = graphics.Viewport.Height / 4;

                int left = graphics.Viewport.X + offsetX;
                int top = graphics.Viewport.Y + offsetY;
                int right = graphics.Viewport.X + graphics.Viewport.Width - offsetX;
                int bottom = graphics.Viewport.Y + graphics.Viewport.Height - offsetY;

                calibrationExpected = new Vector2[] {
                    new Vector2(left, top),
                    new Vector2(right, top),
                    new Vector2(right, bottom),
                    new Vector2(left, bottom)
                };

                calibrationActual = new Vector2[calibrationExpected.Length];

                calibrationIndex = 0;
                UpdateCalibrationAsteroid();
            }

            if (!Calibrating) {
                // Not calibrating, so consider spawning an asteroid.
                if (gameTime.TotalGameTime > nextAsteroidSpawn) {
                    // Too hard:
                    // nextAsteroidSpawn = gameTime.TotalGameTime +
                    //    TimeSpan.FromSeconds(1.0f / Math.Log(6 + score, 6));
                    nextAsteroidSpawn = gameTime.TotalGameTime +
                        TimeSpan.FromSeconds(1.0f / Math.Log(2 + score, 10));

                    Shape asteroid = new Shape(this);
                    asteroid.ModelPath = LargeAsteroidModel;
                    asteroid.TexturePath = LargeAsteroidTexture;
                    asteroid.Position = new Vector3(
                        GameTools.RandomFloat(-10000.0f, +10000.0f), +10000.0f, 0.0f);
                    asteroid.Velocity = new Vector3(
                        GameTools.RandomFloat(-1000.0f, +1000.0f), -1000.0f, 0.0f);
                    asteroid.Acceleration = new Vector3(
                        0.0f, -500f, 0.0f);
                    asteroid.AngularVelocity = Quaternion.CreateFromAxisAngle(
                        GameTools.RandomVector3(),
                        GameTools.RandomFloat(1.0f, 10.0f));

                    asteroid.AmbientLightColor = Vector3.One * 0.15f;
                    asteroid.DiffuseLightColor = Vector3.One * 0.5f;
                    asteroid.SpecularLightColor = Vector3.One * 0.6f;
                    asteroid.Setup();
                    shapes.Add(asteroid);
                }
            }

            List<FtirEvent> ftirEvents = new List<FtirEvent>();

            { // Mouse operations
                MouseState mouseState = Mouse.GetState();

                if (mouseState.LeftButton == ButtonState.Pressed &&
                    oldMouseState.LeftButton == ButtonState.Released) {
                    ftirEvents.Add(new FtirEvent(+1,
                        mouseState.X, mouseState.Y,
                        2 * Settings.Default.BlobMinimumMass));
                } else if (mouseState.LeftButton == ButtonState.Released &&
                    oldMouseState.LeftButton == ButtonState.Pressed) {
                    ftirEvents.Add(new FtirEvent(-1,
                        mouseState.X, mouseState.Y,
                        2 * Settings.Default.BlobMinimumMass));
                }

                oldMouseState = mouseState;
            }

            TimeSpan selfPlayTime = lastEvent + TimeSpan.FromSeconds(30);
            if (selfPlay || selfPlayTime < gameTime.TotalRealTime) {
                int asteroidCount = 0;
                foreach (Shape s in shapes) {
                    if (s.ModelPath == LargeAsteroidModel) {
                        asteroidCount += 1;
                    }
                }

                int count = -1;
                while (GameTools.RandomInt(0, 5) == 0) {
                    count += 1;
                }
                for (int i = 0; i < count; ++i) {
                    int index = GameTools.RandomInt(0, shapes.Count);
                    Vector2 p;
                    if (index < shapes.Count && shapes[index].ModelPath == LargeAsteroidModel) {
                        Vector3 p3 = GraphicsManager.GraphicsDevice.Viewport.Project(
                            shapes[index].Position, Projection, View, Matrix.Identity);
                        p.X = p3.X + GameTools.RandomFloat(-50.0f, 50.0f);
                        p.Y = p3.Y + GameTools.RandomFloat(0.0f, 200.0f);
                    } else {
                        p.X = GameTools.RandomFloat(0.0f, graphics.Viewport.Width);
                        p.Y = GameTools.RandomFloat(0.0f, graphics.Viewport.Height);
                    }

                    if (p.X < 0 || p.X > graphics.Viewport.Width ||
                        p.Y < 0 || p.Y > graphics.Viewport.Height) {
                        // Drop this missile.
                        continue;
                    }

                    FireRocketAt(selfPlayId++, GameTools.UnprojectToZeroPlane(this, p));
                    score = Math.Max(0, score - 1);
                }
            }

            { // FTIR operations
                if (ftir != null) {
                    ftirEvents.AddRange(ftir.PollEvents(Settings.Default.BlobMinimumMass));

                    if (ftirEvents.Count(e => e.Press) > 0) {
                        lastEvent = gameTime.TotalRealTime;
                    }
                }

                if (!showEvents || Calibrating) {
                    particles.Drift = Vector2.Zero;
                } else {
                    particles.Drift = new Vector2(0.0f, -150.0f);

                    float scaleX = graphics.Viewport.Width;
                    float scaleY = graphics.Viewport.Height;
                    float massScale = graphics.Viewport.Width / 20000.0f;

                    // Display dots for the events...
                    foreach (FtirEvent e in ftirEvents) {

                        float eventX = e.Position.X * scaleX;
                        float eventY = e.Position.Y * scaleY;
                        Vector4 color;

                        if (e.Press) {
                            color = new Vector4(0.5f, 1.0f, 0.5f, 1.0f);
                            massScale *= 2;
                        } else if (e.Release) {
                            color = new Vector4(1.0f, 0.5f, 0.5f, 1.0f);
                            massScale *= 2;
                        } else {
                            color = new Vector4(0.4f, 0.4f, 0.8f, 0.5f);
                        }

                        particles.AddFtirEvent(new Vector2(eventX, eventY), massScale * e.Mass, color);
                    }
                }

                if (Calibrating) {
                    // Currently calibrating, want for a single blob with positive mass and
                    // suppress double-taps.
                    if (ftirEvents.Count >= 1 &&
                        ftirEvents[0].Mass > 0 &&
                        calibrationLastEvent + 2.0f < gameTime.TotalGameTime.TotalSeconds) {
                        calibrationActual[calibrationIndex] = ftirEvents[0].Position;
                        FireRocketAt(-1, calibrationAsteroid.Position);
                        calibrationLastEvent = gameTime.TotalGameTime.TotalSeconds;
                    }
                } else if (!showEvents) {
                    // Not calibrating and not showing events, fire a missile at every blob.

                    if (ftirEvents.Count > 0) {
                        foreach (FtirEvent ftirEvent in ftirEvents) {
                            if (ftirEvent.Press) {
                                if (missiles.Count > GameTools.RandomFloat(10, 30)) {
                                    // Unfortunately the (painfully inefficient) engine starts to
                                    // break down with too many rockets. This could definitely be pushed
                                    // farther with a little work, but is not the point of the demo.
                                    break;
                                }

                                //FireLaserAt(GameTools.UnprojectToZeroPlane(this, ftirEvent.Position));
                                Vector2 p = ftirEvent.Position;
                                p.X *= GraphicsDevice.Viewport.Width;
                                p.Y *= GraphicsDevice.Viewport.Height;
                                FireRocketAt(ftirEvent.SessionId, GameTools.UnprojectToZeroPlane(this, p));
                                score = Math.Max(0, score - 1);

                            } else if (ftirEvent.Release) {
                                missiles.Remove(ftirEvent.SessionId);
                            } else {
                                GuidedShape missile;
                                if (missiles.TryGetValue(ftirEvent.SessionId, out missile)) {
                                    if (missile.Deleted) {
                                        missiles.Remove(ftirEvent.SessionId);
                                    } else {
                                        Vector2 p = ftirEvent.Position;
                                        p.X *= GraphicsDevice.Viewport.Width;
                                        p.Y *= GraphicsDevice.Viewport.Height;
                                        missile.Target = GameTools.UnprojectToZeroPlane(this, p);
                                    }
                                }
                            }
                        }
                    }
                }
            }

            { // Keyboard operations

                KeyboardState keyboardState = Keyboard.GetState();

                // Escape ends the game
                if (keyboardState.IsKeyDown(Keys.Escape)) {
                    Exit();
                }

                if (keyboardState.IsKeyDown(Keys.A) && oldKeyboardState.IsKeyUp(Keys.A)) {
                    score += (int)GameTools.RandomFloat(500, 1000);
                }

                if (keyboardState.IsKeyDown(Keys.I) && oldKeyboardState.IsKeyUp(Keys.I)) {
                    showInput = !showInput;
                    if (showInput) {
                        showOutput = false;
                    }
                }
                if (keyboardState.IsKeyDown(Keys.O) && oldKeyboardState.IsKeyUp(Keys.O)) {
                    showOutput = !showOutput;
                    if (showOutput) {
                        showInput = false;
                    }
                }

                if (keyboardState.IsKeyDown(Keys.S) && oldKeyboardState.IsKeyUp(Keys.S)) {
                    if (!selfPlay) {
                        soundBank.PlayCue("calibrate_beep");
                        selfPlay = true;
                    } else {
                        soundBank.PlayCue("calibrate_done");
                        selfPlay = false;
                    }
                }

                // Pressing C starts the calibration
                if (keyboardState.IsKeyDown(Keys.C) && oldKeyboardState.IsKeyUp(Keys.C)) {
                    if (keyboardState.IsKeyDown(Keys.LeftAlt) || keyboardState.IsKeyDown(Keys.RightAlt)) {
                        calibrationIndex = -2;
                        ftir.Calibrate(null, null);
                        UpdateCalibrationAsteroid();
                    } else {
                        calibrationIndex = -1;
                    }
                }

                // Pressing E toggles display of the raw events.
                if (keyboardState.IsKeyDown(Keys.E) && oldKeyboardState.IsKeyUp(Keys.E)) {
                    showEvents = !showEvents;
                }

                // Pressing F toggles between windowed and full screen
                if (keyboardState.IsKeyDown(Keys.F) && oldKeyboardState.IsKeyUp(Keys.F)) {
                    graphicsManager.ToggleFullScreen();
                }

                if (ftir != null) {

                    if (showInput || showOutput) {
                        VideoFormat format = ftir.Format;
                        if (previewTexture == null ||
                            previewTexture.Width != format.Width ||
                            previewTexture.Height != format.Height) {
                            previewTexture = new Texture2D(graphics, format.Width, format.Height,
                                1, TextureUsage.None, SurfaceFormat.Color);
                        }
                        background.VideoFrame = previewTexture;
                    }

                    if (showInput) {
                        ftir.PreviewInput(previewTexture);
                    } else if (showOutput) {
                        ftir.PreviewFiltered(previewTexture);
                    } else {
                        background.VideoFrame = null;
                    }
#if false
                    // Pressing B returns to the background
                    if (keyboardState.IsKeyDown(Keys.B) && oldKeyboardState.IsKeyUp(Keys.B)) {
                        background.SetRunningVideo(-1);
                    }

                    // Pressing D changes the reference frame delay
                    if (keyboardState.IsKeyDown(Keys.D) && oldKeyboardState.IsKeyUp(Keys.D)) {
                        try {
                            ftir.BackgroundDelay += 1;
                        } catch (ArgumentException) {
                            ftir.BackgroundDelay = 0;
                        }
                    }

                    // Pressing S sets or updates a fixed reference frame
                    if (keyboardState.IsKeyDown(Keys.S) && oldKeyboardState.IsKeyUp(Keys.S)) {
                        if (ftir != null) {
                            ftir.BackgroundDelay = 0;
                        }
                    }

                    // Pressing I starts input preview
                    if (keyboardState.IsKeyDown(Keys.I) && oldKeyboardState.IsKeyUp(Keys.I)) {
                        background.SetRunningVideo(0);
                    }

                    // Pressing O starts change preview
                    if (keyboardState.IsKeyDown(Keys.O) && oldKeyboardState.IsKeyUp(Keys.O)) {
                        background.SetRunningVideo(1);
                    }

                    if (keyboardState.IsKeyDown(Keys.OemPlus) && oldKeyboardState.IsKeyUp(Keys.OemPlus)) {
                        ftir.BlobSmoothing += 1;
                    }

                    if (keyboardState.IsKeyDown(Keys.OemMinus) && oldKeyboardState.IsKeyUp(Keys.OemMinus)) {
                        int smoothing = ftir.BlobSmoothing;
                        if (smoothing > 0) {
                            smoothing -= 1;
                        }
                        ftir.BlobSmoothing = smoothing;
                    }
#endif
                }

                oldKeyboardState = keyboardState;
            }

            IsMouseVisible = !graphicsManager.IsFullScreen;

            // Update shapes
            foreach (Shape shape in shapes) {
                shape.Update(gameTime);

                // Delete shapes which fall off the bottom of the field.
                if (shape.Position.Y < -10000) {
                    int grace = 3000;
                    if (shape.Position.X > -10000 + grace && shape.Position.Y < +10000 - grace) {
                        int penalty = Math.Max(1, score / 20);
                        score = Math.Max(0, score - penalty);
                    }
                    shape.Deleted = true;
                }
            }

            // Remove any deleted shapes
            List<int> deletedMissiles = new List<int>();
            foreach (KeyValuePair<int, GuidedShape> entry in missiles)
            {
                if (entry.Value.Deleted)
                {
                    deletedMissiles.Add(entry.Key);
                }
            }
            foreach (int i in deletedMissiles) missiles.Remove(i);

            shapes.RemoveAll(delegate(Shape shape) {
                if (shape.Deleted) {
                    shape.Dispose();
                    return true;
                }
                return false;
            });

            if (audioEngine != null) {
                audioEngine.Update();
            }

            base.Update(gameTime);
        }

        protected override void Draw(GameTime gameTime) {
            GraphicsDevice graphics = graphicsManager.GraphicsDevice;

            // Render content

            graphics.Clear(Color.Black);

            foreach (Shape shape in shapes) {
                bool visible = true;
                if (showEvents && !Calibrating) {
                    visible = shape is Particles || shape is Background;
                }
                if (visible) {
                    shape.Draw(gameTime);
                }
            }

            base.Draw(gameTime);
        }

        public void FireRocketAt(int id, Vector3 target) {
            Vector3 sourceToTarget = target - saucer.Position;
            if (sourceToTarget.Length() > 2000.0f) {
                GuidedShape rocket = new GuidedShape(this);
                rocket.ModelPath = RocketModel;
                rocket.Position = saucer.Position;
                rocket.Velocity = new Vector3(0.0f, 0.0f, 5000.0f);
                rocket.Target = target;
                rocket.MaxAcceleration = 20000.0f;
                rocket.MaxAngularVelocity = MathHelper.ToRadians(360.0f);
                rocket.DiffuseLightColor = Vector3.One * 0.4f;
                rocket.SpecularLightColor = Vector3.One * 0.1f;
                rocket.Lifespan = 5.0f;

                rocket.Arrived += RocketArrived;

                rocket.Setup();
                shapes.Add(rocket);

                if (id >= 0) {
                    missiles[id] = rocket;
                }

                if (soundBank != null) {
                    soundBank.PlayCue("rocket_launch");
                }
            }
        }

        private void RocketArrived(GameTime time, GuidedShape rocket) {
            rocket.Deleted = true;

            foreach (Shape shape in shapes) {
                bool hit = false;
                if (Array.IndexOf(AsteroidModels, shape.ModelPath) >= 0 && shape.Model != null) {
                    foreach (ModelMesh mesh in shape.Model.Meshes) {
                        BoundingSphere meshBounds = mesh.BoundingSphere;
                        Vector3 shapeLocalCenter =
                            Vector3.Transform(meshBounds.Center, mesh.ParentBone.Transform);
                        Vector3 absoluteCenter =
                            Vector3.Transform(shapeLocalCenter, shape.Rotation) + shape.Position;

                        if (Vector3.Distance(rocket.Target, absoluteCenter) < meshBounds.Radius + 500.0f) {
                            hit = true;
                            break;
                        }
                    }
                }

                if (hit) {
                    // Found an asteroid within range of the explosion.
                    soundBank.PlayCue("rock_destroy");

                    if (shape == calibrationAsteroid) {
                        // Move to the next calibration point...
                        calibrationIndex += 1;
                        UpdateCalibrationAsteroid();
                    } else if (shape.ModelPath == LargeAsteroidModel) {
                        score += 10;
                        shape.Deleted = true;

                        particles.AddDust(shape.Position, 100.0f);
                    }
                }
            }

            // Explode at the actual target coordinates rather than on the rocket.
            particles.AddExplosion(rocket.Target, 0.10f * viewportWidth);
            if (soundBank != null) {
                soundBank.PlayCue("rocket_miss");
            }
        }
    }
}
