#include "common.hpp"
#include "blob_detect.hpp"

//EQUIS_USING_NS2(equis,image);

BlobDetector::
BlobDetector()
    : m_threshold(256), m_min_strength(0) {
}

uint16 BlobDetector::
GetThreshold() const {
    return m_threshold;
}

void BlobDetector::
SetThreshold(uint16 value) {
    m_threshold = value;
}

float BlobDetector::
GetMinStrength() const {
    return m_min_strength;
}

void BlobDetector::
SetMinStrength(float value) {
    m_min_strength = value;
}

void BlobDetector::
Update(const array2<uint16>& img) {

    m_stats.EnterFrame();

    const long rows = img.rows();
    const long cols = img.cols();

    /* Flush out the previous state. This leaves m_free and m_row in a
       state equivalent to that following a row with no
       above-threshold pixels. */
    ClearState(rows, cols);

    for (long r = 0; r < rows; ++r) {
        const uint16* row = img.row(r);

        /* Blob of the pixel to the left. At c > 0 this will have been
           initialized at the bottom of the for loop. */
        BlobI* left = NULL;

        for (long c = 0; c < cols; ++c) {
            long mass = row[c];
            
            /* Get the blob of the pixel above this location.
               At y == 0 this will always be NULL, as set by ClearState. 
               'aboveLocal' is used below in garbage collection. */
            BlobI* aboveLocal = m_row[c];
            BlobI* above = GetRoot(aboveLocal);

            BlobI* self;
            if (mass < m_threshold) { /* Not part of any blob. */
                self = NULL;
                
            } else if (left != NULL) { /* Continue the blob to the left. */
                _ASSERT(left->row == r);

                if (above != NULL && above != left) {
                    /* Merge the above blob into the left blob. */
                    above->parent = left;
                    left->x += above->x;
                    left->y += above->y;
					left->brightest = max(left->brightest, above->brightest);
					left->area += above->area;
                    left->mass += above->mass;
                }

                self = left;

            } else if (above != NULL) { /* Continue the blob above. */
                self = above;

            } else { /* Start a new blob. */

                /* As there are only m_width possible blobs in use
                   across rows, the free list should never be
                   emptied. */
                _ASSERT(m_free != NULL);

                self = m_free;
                m_free = m_free->parent;

                /* Quickly initialize the current blob to safe values. */
                memset(self, 0, sizeof(*self));
            }

            if (self) { /* Add the current pixel to its blob. */

                /* 'self' must always be a root blob. */
                _ASSERT(self->parent == NULL);

                self->x += c * mass;
                self->y += r * mass;
				self->brightest = max(self->brightest, static_cast<float>(mass));
                self->mass += mass;
				self->area += 1;

                /* Update the reference count. */
                if (self->row == r) {
                    self->refs += 1;
                } else {
                    /* The reference count applies to the previous row. */
                    self->row = r;
                    self->refs = 1;
                }
            }

            /* Keep the current blob for reference by the pixels to
               the right and below. */
            m_row[c] = self;
            left = self;

            if (aboveLocal && aboveLocal->row != r) {
                /* The blob above must have a reference count greater
                   than zero. Otherwise, it should have been removed
                   when the previous reference was removed. */
                _ASSERT(aboveLocal->refs > 0);

                /* The blob above is defined, but has not been used in
                   this row. This means that the current column
                   discards one of its references. */
                aboveLocal->refs -= 1;

                if (aboveLocal->refs == 0) {
                    /* The reference count of a local blob has dropped
                       to zero, so we must free it. If the local blob
                       was not a root, its values have already been
                       merged. */
                    
                    if (aboveLocal->parent == NULL) {
                        /* The local blob is a root, so represents the
                           final lowest point of the blob. */
                        
                        m_stats.AddSample(aboveLocal->mass);
                        
                        if (aboveLocal->mass >= m_best.back().mass) {
                            Blob save = *aboveLocal;
                            save.x /= save.mass;
                            save.y /= save.mass;
                            m_best.back() = save;

                            HeapBubbleUp();
                        }
                    }

                    /* Push the freed blob onto the free stack. */
                    aboveLocal->parent = m_free;
                    m_free = aboveLocal;
                }
            }
        }
    }

    /* Now check each blob in the final row. To avoid consider the
       same one twice, their counts will be immediately reset to
       zero. These is no need to free these blobs, as ClearState at
       the beginning of the next update will handle it. */
    for (long c = 0; c < cols; ++c) {
        if (m_row[c] != NULL && m_row[c]->refs > 0) {
            /* Zero this blob to ensure we do not add it twice. */
            m_row[c]->refs = 0;

			m_stats.AddSample(m_row[c]->mass);

			if (m_row[c]->mass >= m_best.back().mass) {
                Blob save = *m_row[c];
                save.x /= save.mass;
                save.y /= save.mass;
                m_best.back() = save;

                HeapBubbleUp();
            }
        }
    }

    /* Note that m_best is a reversed heap, rather than completely
       sorted. We defer the sort until the call to Poll, since
       multiple update cycles may be combined. */

    m_stats.LeaveFrame();
}

void BlobDetector::
ClearState(long rows, long cols) {
    (void) rows;

    m_row.resize(cols);
    for (long i = 0; i < cols; ++i) {
        m_row[i] = NULL;
    }

    if (!m_best.size()) {
        m_best.resize(15);
    }

    m_blobs.resize(2 * cols + 2);
    
    /* Build a new free stack from the blobs array. */
    m_free = &m_blobs[0];
    for (long i = 0; i < m_blobs.size() - 1; ++i) {
        m_blobs[i].parent = &m_blobs[i+1];
    }
    m_blobs[m_blobs.size()-1].parent = NULL;
}

void BlobDetector::
HeapBubbleUp() {
    /* This is a classic 1-array-based heap. However, the indices in
       the array are reversed via subtraction from size. */
    int index = 1;    

    const long size = m_best.size();
    while (2 * index + 1 <= size) {
        int leftIndex = 2 * index;
        int rightIndex = 2 * index + 1;
        
        Blob& curr = m_best[size - index];
        Blob& left = m_best[size - leftIndex];
        Blob& right = m_best[size - rightIndex];

        if (left.mass < right.mass) {
            if (left.mass < curr.mass) {
                /* Bubble left. */
                swap(left, curr);
                index = leftIndex;
            } else break;
        } else {
            if (right.mass < curr.mass) {
                /* Bubble right. */
                swap(right, curr);
                index = rightIndex;
            } else break;
        }
    }
}

BlobDetector::BlobI* BlobDetector::
GetRoot(BlobI* blob) {
    if (!blob || !blob->parent) {
        return blob;
    }
    BlobI* root = GetRoot(blob->parent);
    blob->parent = root;
    return root;
}

bool BlobDetector::
IsAncestor(BlobI* ancestor, BlobI* child) {
    for (BlobI* curr = child; curr; curr = curr->parent) {
        if (curr == ancestor) return true;
    }
    return false;
}

struct BlobStrengthLess {
    bool operator () (const Blob& a, const Blob& b) {
        return a.mass > b.mass;
    }
};

long BlobDetector::
Peek(vector<Blob> &values) {
    sort(m_best.begin(), m_best.end(),
         BlobStrengthLess());
    copy(m_best.begin(), m_best.end(),
         back_insert_iterator< vector<Blob> >(values));
    fill(m_best.begin(), m_best.end(),
         Blob());
    return m_best.size();
}
