#include "common.hpp"
#include "blob_track.hpp"

static bool same_position(const Blob& a, const Blob& b) {
	// TODO: should probably use an epsilon on all of these
	return a.x == b.x && a.y == b.y && a.mass == b.mass;
}

BlobTracker::
BlobTracker()
    : m_frame(0),
	  m_next_session_id(1), m_next_cursor_id(0),
      m_min_age(1), m_max_age(1),
      m_step(64) {
}

BlobTracker::
~BlobTracker() {
}

long BlobTracker::
GetMinAge() {
    return m_min_age;
}

void BlobTracker::
SetMinAge(long value) {
    m_min_age = value;
}

long BlobTracker::
GetMaxAge() {
    return m_max_age;
}

void BlobTracker::
SetMaxAge(long value) {
    m_max_age = value;
}

void BlobTracker::
Resize(long rows, long cols) {
	m_rows = rows;
	m_cols = cols;

    long crows = (rows + m_step - 1) / m_step;
    long ccols = (cols + m_step - 1) / m_step;
    m_cells.resize(crows, ccols);
}

#define NEVER_MATCH (1.0e9f)

float BlobTracker::
MatchPenalty(const BlobI& curr, const BlobI& prev) {
    if (curr.last_frame != prev.last_frame + 1) return NEVER_MATCH;
    
    float dx = fabs(curr.x - prev.x);
    float dy = fabs(curr.y - prev.y);
    float distance = sqrt(dx * dx + dy * dy);

    return distance;
}

BlobTracker::BlobI* BlobTracker::
FindBestMatch(const BlobI& curr) {
    long cr = static_cast<long>(curr.y / m_step);
    long cc = static_cast<long>(curr.x / m_step);
    
    float best_penalty = 128;
    BlobI *best = NULL;

    /* Search the surrounding cells... */
    for (int dr = -1; dr <= +1; ++dr) {
        for (int dc = -1; dc <= +1; ++dc) {
            if (!m_cells.valid(cr + dr, cc + dc)) continue;
            
            Cell& cell = m_cells(cr + dr, cc + dc);
            for (size_t j = 0; j < cell.blobs.size(); ++j) {
                float penalty = MatchPenalty(curr, cell.blobs[j]);
                if (penalty < best_penalty) {
                    best_penalty = penalty;
                    best = &cell.blobs[j];
                }
            }
        }
    }

    return best;
}

void BlobTracker::
Update(const vector<Blob>& blobs) {
    m_frame += 1;
    
    for (size_t i = 0; i < blobs.size(); ++i) {
        BlobI curr(blobs[i]);
		curr.first_frame = m_frame;
        curr.last_frame = m_frame;

        BlobI* match = FindBestMatch(curr);
        if (match == NULL) {

            /* This blob might become a touch later, in which case it
               will be allocated non-negative ids. */
            curr.cursor_id = -1;
            curr.session_id = -1;

            long cr = static_cast<long>(curr.y / m_step);
            long cc = static_cast<long>(curr.x / m_step);

            if (m_min_age == 0) {
                /* No aging required to become a touch. */
				curr.cursor_id = create_cursor_id();
                curr.session_id = m_next_session_id++;
                m_events.push_back(Event(1, m_frame, m_frame, curr));
            }
            
            m_cells(cr, cc).blobs.push_back(curr);
        } else {
            /* This blob contributes to a previous touch. */
            curr.cursor_id = match->cursor_id;
            curr.session_id = match->session_id;
			curr.first_frame = match->first_frame;

			if (curr.session_id < 0) {
				/* This blob has not yet become a touch. */
				if (curr.first_frame + m_min_age <= curr.last_frame) {
					/* Now a touch. Allocate an id and report the event. */
					curr.cursor_id = create_cursor_id();
					curr.session_id = m_next_session_id++;
					m_events.push_back(Event(1, m_frame, curr.first_frame, curr));
				}
			} else if (!same_position(curr, *match)) {
				m_events.push_back(Event(0, m_frame, curr.first_frame, curr));
			}
            
            long cr = static_cast<long>(curr.y / m_step);
            long cc = static_cast<long>(curr.x / m_step);
            long match_cr = static_cast<long>(match->y / m_step);
            long match_cc = static_cast<long>(match->x / m_step);
            
            if (cr == match_cr && cc == match_cc) {
                *match = curr;
            } else {
                m_cells(cr, cc).blobs.push_back(curr);

                /* Moving between cells; delete the old entry below. */
                match->last_frame = -1;
            }
        }
    }

    /* Cull old blobs. */
    for (long r = 0; r < m_cells.rows(); ++r) {
        for (long c = 0; c < m_cells.cols(); ++c) {
            Cell& cell = m_cells(r, c);

            vector<BlobI>::iterator read = cell.blobs.begin();
            vector<BlobI>::iterator write = cell.blobs.begin();

            for (; read != cell.blobs.end(); ++read) {
                if (read->last_frame < 0) {
                    // This blob was flagged as deleted above.
					// Do not deallocate the cursor id.
                } else if (read->last_frame + m_max_age < m_frame) {
                    // This touch has not had a blob in too long, so
                    // discard it. Only report the event if the
                    // corresponding add has already been sent.
                    if (read->session_id >= 0) {
                        m_events.push_back(Event(-1, m_frame, read->first_frame, *read));
						release_cursor_id(read->cursor_id);
                    }
                } else {
                    if (read != write) *write = *read;
                    write += 1;
                }
            }
            
            cell.blobs.erase(write, cell.blobs.end());
        }
    }
}

void BlobTracker::
Poll(vector<Event>& events) {
    copy(m_events.begin(), m_events.end(),
         back_insert_iterator< vector<Event> >(events));
    m_events.clear();
}

long BlobTracker::
Poll(ETEvent* events, long events_cap) {
	long count = min(m_events.size(), (size_t)events_cap);
	for (long i = 0; i < count; ++i) {
		events[i].frame = m_events[i].frame;
		events[i].since = m_events[i].since;
		events[i].type = m_events[i].type;
		events[i].cursor_id = m_events[i].blob.cursor_id;
		events[i].session_id = m_events[i].blob.session_id;
		events[i].x = m_events[i].blob.x / m_cols;
		events[i].y = m_events[i].blob.y / m_rows;
		events[i].mass = m_events[i].blob.mass;
	}
    if (count) m_events.erase(m_events.begin(), m_events.begin() + count);
	return count;
}

long BlobTracker::
create_cursor_id() {
	set<long>::iterator it = m_free_cursor_ids.begin();
	if (it == m_free_cursor_ids.end()) {
		return m_next_cursor_id++;
	} else {
		long result = *it;
		m_free_cursor_ids.erase(it);
		return result;
	}
}

void BlobTracker::
release_cursor_id(long cursor_id) {
	m_free_cursor_ids.insert(cursor_id);
}

BlobTracker::BlobI::
BlobI(const Blob& source) : Blob(source) {
}
