#include "common.hpp"
#include "calibration.hpp"
EQUIS_USING_NS1(equis);

#include <shlobj.h>
#include <shlwapi.h>

#include <vector>

float distance_sq(const vec2f& a, const vec2f& b) {
	float dx = a.x - b.x;
	float dy = a.y - b.y;
	return dx * dx + dy * dy;
}

float distance(const vec2f& a, const vec2f& b) {
	return sqrt(distance_sq(a, b));
}

float dot(const vec2f& a, const vec2f& b) {
	return a.x * b.x + a.y * b.y;
}

float cross(const vec2f& a, const vec2f& b) {
	return a.y * b.x - a.x * b.y;
}

float magnitude(const vec2f& a) {
	return sqrt(a.x * a.x + a.y * a.y);
}

vec2f operator + (const vec2f& a, const vec2f& b) {
	vec2f r;
	r.x = a.x + b.x;
	r.y = a.y + b.y;
	return r;
}

vec2f operator - (const vec2f& a, const vec2f& b) {
	vec2f r;
	r.x = a.x - b.x;
	r.y = a.y - b.y;
	return r;
}

vec2f operator * (const vec2f& a, const float& b) {
	vec2f r;
	r.x = a.x * b;
	r.y = a.y * b;
	return r;
}

Calibration::Calibration() {
	load();
}

void Calibration::load() {
	HRESULT res;
	WCHAR path[MAX_PATH];

	res = SHGetFolderPath(NULL,
		CSIDL_APPDATA,
		NULL, SHGFP_TYPE_CURRENT,
		path);
	if (res != S_OK) return;

	wcscat_s(path, L"\\Equis");
	if (GetFileAttributes(path) == INVALID_FILE_ATTRIBUTES) {
		if (!CreateDirectory(path, NULL)) return;
	}

	wcscat_s(path, L"\\calibration.txt");

	// Assume that any failure means the file does not exist.
	if (GetFileAttributes(path) == INVALID_FILE_ATTRIBUTES) return;

	FILE* in = _wfopen(path, L"rt");
	if (in == NULL) return;

	int rows, cols;
	if (fscanf(in, "%d%d", &rows, &cols) != 2) goto end;

	m_table.resize(rows, cols);
	for (int r = 0; r < rows; ++r) {
		for (int c = 0; c < cols; ++c) {
			float x, y;
			if (fscanf(in, " (%f ,%f )", &x, &y) != 2) goto end;
			m_table(r, c).x = x;
			m_table(r, c).y = y;
		}
	}

#if 0
	for (size_t row = 0; row < m_table.rows(); ++row) {
		for (size_t col = 0; col < m_table.cols(); ++col) {
			printf("(%0.2f,%0.2f) ", m_table(row,col).x, m_table(row,col).y);
		}
		printf("\n");
	}
#endif

end:
	fclose(in);
}

void Calibration::save() {
	HRESULT res;
	WCHAR path[MAX_PATH];

	res = SHGetFolderPath(NULL,
		CSIDL_APPDATA | CSIDL_FLAG_CREATE,
		NULL, SHGFP_TYPE_CURRENT,
		path);
	if (res != S_OK) return;

	wcscat_s(path, L"\\Equis");
	if (GetFileAttributes(path) == INVALID_FILE_ATTRIBUTES) {
		if (!CreateDirectory(path, NULL)) return;
	}

	wcscat_s(path, L"\\calibration.txt");
	printf("Saving %ws\n", path);

	FILE* out = _wfopen(path, L"wt");
	if (out == NULL) return;

	int rows = m_table.rows(), cols = m_table.cols();
	fprintf(out, "%d %d\n", rows, cols);

	for (int r = 0; r < rows; ++r) {
		for (int c = 0; c < cols; ++c) {
			if (c != 0) fputc(' ', out);
			fprintf(out, "(%f,%f)", m_table(r, c).x, m_table(r, c).y);
		}
		fputc('\n', out);
	}

	fclose(out);
}

void
Calibration::clear() {
	for (long row = 0; row < m_table.rows(); ++row) {
		for (long col = 0; col < m_table.cols(); ++col) {
			vec2f curr;
			curr.x = col * 1.0f / (m_table.cols() - 1);
			curr.y = row * 1.0f / (m_table.rows() - 1);

			m_table(row, col) = curr;
		}
	}
}

bool
Calibration::update(const array<vec2f>& expected, const array<vec2f>& detected) {

	m_table.resize(8,8);//640/8, 480/8);
	m_table.clear();

	size_t count = min(expected.size(), detected.size());
	if (count == 0) return false;

	vector< pair<vec2f, vec2f> > points;
	for (size_t i = 0; i < count; ++i) {
		points.push_back(make_pair(detected[i], expected[i]));
	}

	vector< pair<float, int> > points_by_distance;
	for (size_t i = 0; i < points.size(); ++i) {
		points_by_distance.push_back(make_pair(0.0f, i));
	}

	for (long row = 0; row < m_table.rows(); ++row) {
		for (long col = 0; col < m_table.cols(); ++col) {
			vec2f curr;
			curr.x = col * 1.0f / (m_table.cols() - 1);
			curr.y = row * 1.0f / (m_table.rows() - 1);

			for (size_t i = 0; i < points.size(); ++i) {
				int index = points_by_distance[i].second;
				points_by_distance[i] = make_pair(
					::distance(curr, points[index].first),
					index);
			}
			sort(points_by_distance.begin(), points_by_distance.end());

			// Select the nearest three non-colinear points.
			pair<vec2f, vec2f> ref[3];
			ref[0] = points[points_by_distance[0].second];

			bool found = false;
			for (size_t i = 1; i < points.size(); ++i) {
				for (size_t j = i + 1; j < points.size(); ++j) {
					ref[1] = points[points_by_distance[i].second];
					ref[2] = points[points_by_distance[j].second];

					vec2f r01(ref[1].first - ref[0].first);
					vec2f r02(ref[2].first - ref[0].first);
					if (abs(cross(r01, r02)) > 0.5 * magnitude(r01) * magnitude(r02)) {
						found = true;
						break;
					}
				}
			}

			// Unable to find suitable coordinates.
			if (!found) {
				return false;
			}

			float a, b;
			{
				vec2f r(ref[0].first);
				vec2f p(curr - r);
				vec2f i(ref[1].first - r);
				vec2f j(ref[2].first - r);

				a = cross(j, p) / cross(j, i);
				b = cross(i, p) / cross(i, j);
			}

			m_table(row, col) = ref[0].second +
				(ref[1].second - ref[0].second) * a +
				(ref[2].second - ref[0].second) * b;
		}
	}

#if 0
	for (size_t row = 0; row < m_table.rows(); ++row) {
		for (size_t col = 0; col < m_table.cols(); ++col) {
			printf("(%0.2f,%0.2f) ", m_table(row,col).x, m_table(row,col).y);
		}
		printf("\n");
	}
#endif

	save();

	return true;
}

vec2f
Calibration::transform(vec2f src) {
	if (m_table.empty()) return src;

	size_t row = static_cast<size_t>(src.y * (m_table.rows() - 1));
	size_t col = static_cast<size_t>(src.x * (m_table.cols() - 1));

	float fx = src.x * (m_table.cols() - 1) - col;
	float fy = src.y * (m_table.rows() - 1) - row;

	vec2f nw = m_table(row + 0, col + 0);
	vec2f ne = m_table(row + 0, col + 1);
	vec2f sw = m_table(row + 1, col + 0);
	vec2f se = m_table(row + 1, col + 1);

	return
		(nw * (1 - fx) + ne * fx) * (1 - fy) +
		(sw * (1 - fx) + se * fx) * fy;
}

void
Calibration::transform(array<vec2f>& dst, const array<vec2f>& src) {
	dst.resize(src.size());
	for (long i = 0; i < src.size(); ++i) {
		dst[i] = transform(src[i]);
	}
}

void
Calibration::transform(array<Blob>& blobs) {
	for (long i = 0; i < blobs.size(); ++i) {
		vec2f v;
		v.x = blobs[i].x;
		v.y = blobs[i].y;
		v = transform(v);
		blobs[i].x = v.x;
		blobs[i].y = v.y;
	}
}
