#include "common.hpp"
#include "DSInput.hpp"
#include "DSInputSink.hpp"
EQUIS_USING_NS1(equis);

DSInput::DSInput(const EVInputSpi* spi, IBaseFilter* source, HRESULT& res)
: m_lock(res), m_source(source), m_running(false)
{
	this->spi = spi;

	if (res == S_OK && source == NULL) {
		res = E_POINTER;
	}

	if (res == S_OK) {
		res = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER,
			m_graph.GetUuid(), m_graph.DetachAndGetVoidPointer());
	}

	if (res == S_OK) {
		res = m_graph->QueryInterface(
			m_control.GetUuid(), m_control.DetachAndGetVoidPointer());
	}

	if (res == S_OK) {
		res = m_graph->AddFilter(m_source, L"Source");
	}

	if (res == S_OK) {
		m_sink.AttachWithoutAddRef(new DSInputSink(this, res));
	}

	if (res == S_OK) {
		m_graph->AddFilter(m_sink, L"Sink");
	}

	ComPtr<ICaptureGraphBuilder2> builder;
	if (res == S_OK) {
		res = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC_SERVER,
			builder.GetUuid(), builder.DetachAndGetVoidPointer());
	}
	if (res == S_OK) {
		res = builder->SetFiltergraph(m_graph);
	}

	if (res == S_OK) {
		res = builder->FindPin(m_source, PINDIR_OUTPUT,
			&PIN_CATEGORY_CAPTURE,
			&WMMEDIATYPE_Video,
			TRUE, 0, m_source_pin.DetachAndGetPointer());
		_ASSERT(res == S_OK);
	}

	if (res == S_OK) {
		// The sink's input will not have a type defined yet,
		// so can not restrict it to video.
		res = builder->FindPin(implicit_cast<BaseUnknown*>(m_sink), PINDIR_INPUT,
			NULL, NULL,
			TRUE, 0, m_sink_pin.DetachAndGetPointer());
		_ASSERT(res == S_OK);
	}

	_ASSERT(res == S_OK);
}

DSInput::DSInput(const EVInputSpi* spi, const char* path, HRESULT& res)
	: m_lock(res), m_running(false)
{
	this->spi = spi;

	if (res == S_OK && path == NULL) {
		res = E_POINTER;
	}

	if (res == S_OK) {
		res = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER,
			m_graph.GetUuid(), m_graph.DetachAndGetVoidPointer());
	}

	if (res == S_OK) {
		res = m_graph->QueryInterface(
			m_control.GetUuid(), m_control.DetachAndGetVoidPointer());
	}

	wchar_t wpath[1024];
	if (res == S_OK) {
		if (MultiByteToWideChar(CP_UTF8, 0, path, -1,
				wpath, EQUIS_ARRAY_CAPACITY(wpath)) == 0) {
			res = HRESULT_FROM_WIN32(GetLastError());
		}
	}
	if (res == S_OK) {
		res = m_graph->AddSourceFilter(wpath, L"Source", m_source.DetachAndGetPointer());
	}

	if (res == S_OK) {
		m_sink.AttachWithoutAddRef(new DSInputSink(this, res));
	}

	if (res == S_OK) {
		m_graph->AddFilter(m_sink, L"Sink");
	}

	ComPtr<ICaptureGraphBuilder2> builder;
	if (res == S_OK) {
		res = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC_SERVER,
			builder.GetUuid(), builder.DetachAndGetVoidPointer());
	}
	if (res == S_OK) {
		res = builder->SetFiltergraph(m_graph);
	}

	if (res == S_OK) {
		res = builder->FindPin(m_source, PINDIR_OUTPUT,
			NULL,
			NULL/*&WMMEDIATYPE_Video*/,
			TRUE, 0, m_source_pin.DetachAndGetPointer());
		_ASSERT(res == S_OK);
	}

	if (res == S_OK) {
		ComPtr<IAMStreamControl> ap;
		HRESULT a = builder->FindInterface(NULL, NULL, m_source, ap.GetUuid(), ap.DetachAndGetVoidPointer());

		int x = 0;
	}

	if (res == S_OK) {
		// The sink's input will not have a type defined yet,
		// so can not restrict it to video.
		res = builder->FindPin(implicit_cast<BaseUnknown*>(m_sink), PINDIR_INPUT,
			NULL, NULL,
			TRUE, 0, m_sink_pin.DetachAndGetPointer());
		_ASSERT(res == S_OK);
	}

	_ASSERT(res == S_OK);
}


DSInput::~DSInput() {
	if (m_control.IsNotNull()) {
		m_control->Stop();
	}
}

int DSInput::Close() {
	HRESULT res = S_OK;

	{ // Ensure the critical section is released before deleting 'this'.
		MutexLock lock(&m_lock);
		if (res == S_OK) {
			res = lock.Enter(INFINITE);
		}
	
		if (res == S_OK) {
			m_source_pin.Detach();
			m_source.Detach();
		}

		if (res == S_OK && m_control.IsNotNull()) {
			res = m_control->Stop();
			if (res == S_FALSE) {
				OAFilterState state;
				res = m_control->GetState(INFINITE, &state);
				if (res == S_OK && state != State_Stopped) {
					res = E_FAIL;
				}
			}
		}
	}

	if (res != S_OK) return -1;

	delete this;
	return 0;
}

long DSInput::GetFormats(long start, EVFormat* formats, int formats_cap) {
	HRESULT res = S_OK;

	if (formats == NULL) {
		res = E_POINTER;
	}

	ComPtr<IEnumMediaTypes> types;
	if (res == S_OK) {
		res = m_source_pin->EnumMediaTypes(types.DetachAndGetPointer());
	}

	int formats_end = 0;

	while (res == S_OK && formats_end < formats_cap) {
		MediaTypePtr type;
		res = types->Next(1, type.ClearAndGetPointer(), NULL);
		if (res != S_OK) {
			if (res == S_FALSE) res = S_OK;
			break;
		}

		res = MediaTypeToFormat(&formats[formats_end], type);
		if (res != S_OK) {
			if (res == S_FALSE) res = S_OK;
			continue;
		}

		formats_end += 1;
	}

	_ASSERT(res == S_OK);
	return res == S_OK ? formats_end : -1;
}

int DSInput::SetFormat(EVFormat* format) {

	equis::array<EVFormat> candidates;
	if (!candidates.resize(16)) return -1;

	long count = GetFormats(0, candidates.begin(), candidates.size());
	while (count == candidates.size()) {
		if (!candidates.resize(2 * candidates.size())) return -1;
		count = GetFormats(count, candidates.begin() + count, candidates.size() - count);
	}
	if (count < 0) return -1; // Failed to enumerate formats.

	// If no format was specified, use an empty object instead.
	EVFormat default_format = { 0, 0, 0, 0, 0 };
	if (format == NULL) format = &default_format;

	long index = evfChoose(format, candidates.begin(), count);
	if (index < 0) return -1; // Failed to find a compatible format.
	
	HRESULT res = S_OK;

	{ // Fetch the media type corresponding to the selected format.
		ComPtr<IEnumMediaTypes> types;
		if (res == S_OK) {
			res = m_source_pin->EnumMediaTypes(types.DetachAndGetPointer());
		}
		if (res == S_OK && index > 0) {
			res = types->Skip(index);
		}
		if (res == S_OK) {
			res = types->Next(1, m_source_type.ClearAndGetPointer(), NULL);
		}
	}

	if (res != S_OK) return -1;
	m_actual_format = *format;
	return 0;
}

int DSInput::GetFormat(EVFormat* format) {
	if (!format) return -1;

	if (m_source_type.IsNull()) {
		// Have not yet chosen a type for this filter. Do that now.
		if (SetFormat(NULL) < 0) return -1;
	}

	memcpy(format, &m_actual_format, sizeof(m_actual_format));
	return 0;
}

EVInputFrame* DSInput::Poll() {
	HRESULT res = S_OK;

	MutexLock lock(&m_lock);
	if (res == S_OK) {
		res = lock.Enter(INFINITE);
	}

	if (res == S_OK && !m_running) {
		res = InternalStart();
	}

	DSInputFrame* ds_frame = NULL;
	if (res == S_OK) {
		// Not quite a zero delay, but ensure that we get a frame quickly.
		// Should probably set this up to avoid acquiring the critical section.
		res = m_sink->ReadFrame(5, &ds_frame);
		return ds_frame;
	}
	return ds_frame;
}

EVInputFrame* DSInput::Next() {
	HRESULT res = S_OK;

	MutexLock lock(&m_lock);
	if (res == S_OK) {
		res = lock.Enter(INFINITE);
	}

	if (res == S_OK && !m_running) {
		res = InternalStart();
	}

	DSInputFrame* ds_frame = NULL;
	if (res == S_OK) {
		res = m_sink->ReadFrame(INFINITE, &ds_frame);
		return ds_frame;
	}
	return ds_frame;
}

int DSInput::Callback(EVICallbackProc callback, void* user) {
	HRESULT res = S_OK;

	MutexLock lock(&m_lock);
	if (res == S_OK) {
		res = lock.Enter(INFINITE);
	}

	if (res == S_OK) {
		res = m_sink->SetCallback(callback, user);
	}

	if (res == S_OK) {
		if (!m_running && callback) {
			res = InternalStart();
		}
		else if (m_running && !callback) {
			res = m_control->Stop();
			if (res == S_FALSE) {
				// TODO: Should this wait for the graph to stop?
				//       Probably more intuitive for the caller if this blocks until
				//       after the last invocation of the callback.
				OAFilterState state;
				res = m_control->GetState(INFINITE, &state);
				if (res == S_OK && state != State_Stopped) {
					res = E_FAIL;
				}
			}
		}
	}

	_ASSERT(res == S_OK);
	return res == S_OK ? 0 : -1;
}

int DSInput::Done(EVInputFrame* frame) {
	if (frame == NULL) return E_POINTER;

	HRESULT res = S_OK;

	res = m_sink->
		PushUnusedFrame(
		static_cast<DSInputFrame*>(frame));

	_ASSERT(res == S_OK);
	return res == S_OK ? 0 : -1;
}

int DSInput::Geti(int name, int* value) {
	return -1;
}

int DSInput::Getf(int name, float* value) {
	return -1;
}

int DSInput::Seti(int name, int value) {
	return -1;
}

int DSInput::Setf(int name, float value) {
	return -1;
}

int DSInput::GetCapsf(int name, float* min, float* max, float* step, float* def) {
	return -1;
}

int DSInput::GetStopsf(int name, float* stops, int stops_cap) {
	return -1;
}

HRESULT DSInput::InternalStart() {
	// This function may only be called when m_lock is held.

	if (m_running ||
		m_graph.IsNull() || m_control.IsNull() ||
		m_source.IsNull() || m_source_pin.IsNull() ||
		m_sink.IsNull()) {
		return E_UNEXPECTED;
	}

	HRESULT res = S_OK;

	if (res == S_OK) {
		res = m_graph->Disconnect(m_source_pin);
		if (res == S_FALSE) res = S_OK; // Pin was not connected; ignore.
	}

	if (res == S_OK) {
		res = m_graph->Disconnect(m_sink_pin);
		if (res == S_FALSE) res = S_OK; // Pin was not connected; ignore.
	}

	if (m_source_type.IsNull()) {
		// Have not yet chosen a media type for the source. Do that now.
		if (SetFormat(NULL) < 0) return -1;
	}

	if (res == S_OK) {
		res = m_sink->SetDesired(&m_actual_format, m_source_type);
	}

	if (res == S_OK) {
		res = m_graph->Connect(m_source_pin, m_sink_pin);
	}

	if (res == S_OK) {
		res = m_control->Run();
		if (res == S_FALSE) {
			OAFilterState state;
			res = m_control->GetState(INFINITE, &state);
			if (res == S_OK && state != State_Stopped) {
				res = E_FAIL;
			}
		}
	}

	if (res == S_OK) {
		m_running = true;
	}

	_ASSERT(res == S_OK);
	return res;
}

HRESULT MediaTypeToFormat(EVFormat* dst, const AM_MEDIA_TYPE* src) {

	if (!IsEqualGUID(src->majortype, WMMEDIATYPE_Video) ||
		!IsEqualGUID(src->formattype, WMFORMAT_VideoInfo) ||
		!src->bFixedSizeSamples ||
		src->bTemporalCompression ||
		src->cbFormat < sizeof(VIDEOINFOHEADER) ||
		!src->pbFormat)
	{
		// This media type can not be converted.
		return S_FALSE;
	}

	VIDEOINFOHEADER* srcv = reinterpret_cast<VIDEOINFOHEADER*>(src->pbFormat);
	BITMAPINFOHEADER* srcb = &srcv->bmiHeader;

	memset(dst, 0, sizeof(*dst));

	if (IsEqualGUID(src->subtype, WMMEDIASUBTYPE_RGB32)) dst->types = EVF_TYPE_A8R8G8B8;
	else if (IsEqualGUID(src->subtype, WMMEDIASUBTYPE_Y800)) dst->types = EVF_TYPE_Y8;
	else if (IsEqualGUID(src->subtype, WMMEDIASUBTYPE_Y160)) dst->types = EVF_TYPE_Y16;
	else if (IsEqualGUID(src->subtype, WMMEDIASUBTYPE_I420)) dst->types = EVF_TYPE_COLOR | EVF_TYPE_GRAY;
	else if (IsEqualGUID(src->subtype, WMMEDIASUBTYPE_RGB24)) dst->types = EVF_TYPE_COLOR;
	else if (IsEqualGUID(src->subtype, WMMEDIASUBTYPE_RGB555)) dst->types = EVF_TYPE_COLOR;
	else if (IsEqualGUID(src->subtype, WMMEDIASUBTYPE_RGB565)) dst->types = EVF_TYPE_COLOR;
	else {
		// Unable to convert this media type.
		
		char name[8] = "XXXX ";
		memcpy(name, &src->subtype.Data1, 4);
		for (int i = 0; i < 4; ++i) {
			if (!isprint(name[i])) {
				name[i + 0] = ' ';
				name[i + 1] = '\0';
				break;
			}
		}

		fprintf(stderr,
			"Warning: Ignoring media type due to unknown subtype.\n"
			"  Subtype is %s{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}\n\n",
			name,
			src->subtype.Data1, src->subtype.Data2, src->subtype.Data3,
			src->subtype.Data4[0], src->subtype.Data4[1], src->subtype.Data4[2], src->subtype.Data4[3],
			src->subtype.Data4[4], src->subtype.Data4[5], src->subtype.Data4[6], src->subtype.Data4[7]);
		return S_FALSE;
	}

	dst->width = srcb->biWidth;
	dst->height = srcb->biHeight;
	dst->usecs = static_cast<long>(srcv->AvgTimePerFrame / 10);

	if (dst->height < 0) {
		dst->flags |= EVF_FLAG_BOTTOM_UP;
		dst->height = -dst->height;
	}

	return S_OK;
}
