#include "common.hpp"
#include "DSInput.hpp"
#include "DSInputFrame.hpp"
#include "video/input_spi.h"
#include "Utility/EnumVideoCaptures.hpp"
EQUIS_USING_NS1(equis);

static int dshow_eviEnum(EVInputSpi* self, int start, char* uris, int uris_cap);
static EVInput* dshow_eviOpen(EVInputSpi* self, const char* uri);
static EVInput* dshow_eviOpenFile(EVInputSpi* self, const char* path);

static IBaseFilter* getFilterByPath(const char* path);
static long wcstouri(char* dst, long dst_cap, const wchar_t* src);

static int dshow_Close(EVInput* self);
static int dshow_GetFormats(EVInput* self, int start, EVFormat* formats, int formats_cap);
static int dshow_SetFormat(EVInput* self, EVFormat* format);
static int dshow_GetFormat(EVInput* self, EVFormat* format);
static EVInputFrame* dshow_Poll(EVInput* self);
static EVInputFrame* dshow_Next(EVInput* self);
static int dshow_Callback(EVInput* self, EVICallbackProc callback, void* user);

static const EVFormat* dshow_FrameFormat(EVInputFrame* frame);
static long dshow_FrameSize(EVInputFrame* frame);
static void* dshow_FramePointer(EVInputFrame* frame);
static int dshow_FrameDone(EVInputFrame* frame);

static int dshow_Geti(EVInput* self, int name, int* value);
static int dshow_Getf(EVInput* self, int name, float* value);
static int dshow_Seti(EVInput* self, int name, int value);
static int dshow_Setf(EVInput* self, int name, float value);
static int dshow_GetCapsf(EVInput* self, int name, float* min, float* max, float* step, float* def);
static int dshow_GetStopsf(EVInput* self, int name, float* stops, int stops_cap);

static EVInputSpi s_dshow_spi = {
	dshow_eviEnum,
	dshow_eviOpen,
	dshow_eviOpenFile,

	dshow_Close,
	dshow_GetFormats,
	dshow_SetFormat,
	dshow_GetFormat,
	dshow_Poll,
	dshow_Next,
	dshow_Callback,

	dshow_FrameFormat,
	dshow_FrameSize,
	dshow_FramePointer,
	dshow_FrameDone,

	dshow_Geti,
	dshow_Getf,
	dshow_Seti,
	dshow_Setf,
	dshow_GetCapsf,
	dshow_GetStopsf,
};

void evi_dshow_register_spi() {
	eviRegisterSpi(&s_dshow_spi);
}

static int dshow_eviEnum(EVInputSpi* self, int start, char* uris, int uris_cap) {
	wchar_t index_path[16];

	HRESULT res = S_OK;

	ComPtr<EnumVideoCaptures> captures;
	if (res == S_OK) {
		captures.AttachWithoutAddRef(new EnumVideoCaptures(res));
	}

	int count = 0;
	int uris_end = 0;
	while (res == S_OK && uris_end + 16 < uris_cap) {
		res = captures->Next();
		if (res == S_FALSE) { res = S_OK; break; }

		count += 1;

		// Skip leading entries as directed by start.
		if (start > 0) {
			start -= 1;
			continue;
		}

		LPCWSTR path = NULL; // Memory managed by 'captures'.
		if (res == S_OK) {
			res = captures->GetPath(&path);
			if (res == E_INVALIDARG) {
				// This device does not supply a path.
				// Create a synthetic one based on the current index.
				path = index_path;
				res = StringCbPrintfW(index_path, sizeof(index_path), L"%d", count);
			}
		}

		int path_start = uris_end + 6;
		int wrote = wcstouri(uris + path_start, uris_cap - 1 - path_start, path);
		if (wrote <= 0) {
			DWORD err = GetLastError();
			// Reaching the end of the buffer is not reported.
			if (err != ERROR_INSUFFICIENT_BUFFER) {
				res = HRESULT_FROM_WIN32(err);
			}
			break;
		}

		memcpy(uris + uris_end, "dshow:", 6);
		uris_end = path_start + wrote + 1; // Advance past the URI and NUL.
	}

	if (uris_end < uris_cap) {
		uris[uris_end] = '\0';
	}
	else {
		// Unable to terminate the list.
		res = E_FAIL;
	}

	_ASSERT(res == S_OK);
	return SUCCEEDED(res) ? count : -1;
}

static EVInput* dshow_eviOpen(EVInputSpi* self, const char* uri) {

	// Reject non-DirectShow URIs.
	if (uri != NULL && strncmp(uri, "dshow:", 6) != 0) return NULL;

	ComPtr<IBaseFilter> filter;
	if (uri) {
		const char* path = uri + 6;
		filter.AttachWithoutAddRef(getFilterByPath(path));
	} else {
		filter.AttachWithoutAddRef(getFilterByPath(NULL));
	}

	if (filter.IsNull()) return NULL;

	HRESULT res = S_OK;

	DSInput* input = new DSInput(self, filter, res);
	if (res != S_OK) {
		delete input;
		return NULL;
	}
	return input;
}

static EVInput* dshow_eviOpenFile(EVInputSpi* self, const char* path) {

	if (path == NULL) return NULL;

	HRESULT res = S_OK;

	DSInput* input = new DSInput(self, path, res);
	if (res != S_OK) {
		delete input;
		return NULL;
	}
	return input;
}

static IBaseFilter* getFilterByPath(const char* path) {
	HRESULT res = S_OK;

	long path_index = 1;
	if (path) {
		char* end;
		path_index = strtol(path, &end, 10); 
		if (*end) path_index = -1;
	}

	ComPtr<EnumVideoCaptures> captures;
	if (res == S_OK) {
		captures.AttachWithoutAddRef(new EnumVideoCaptures(res));
	}

	array<char> candidate_uri;

	IBaseFilter* found = NULL;
	int count = 0;
	while (res == S_OK) {
		res = captures->Next();
		if (res == S_FALSE) break;

		count += 1;

		if (path_index > 0) {
			if (path_index == count) {
				res = captures->BindToObject(IID_IBaseFilter, void_cast(&found));
				break;
			}
			else {
				continue;
			}
		}

		LPCWSTR candidate = NULL; // Memory managed by 'captures'.
		if (res == S_OK) {
			res = captures->GetPath(&candidate);
			if (res == E_INVALIDARG) {
				// This device does not supply a path. Have already checked the index,
				// so skip this device.
				res = S_OK;
				continue;
			}
		}

		if (res == S_OK) {
			// In the worst case, a single wchar_t might require three bytes to encode
			// in UTF-8. Since these bytes need to be escaped in the URI, they require
			// a total of nine bytes. Add one byte for the NUL terminator.
			size_t x = wcslen(candidate) * 9 + 1;
			candidate_uri.resize(wcslen(candidate) * 12 + 1);
			if (wcstouri(candidate_uri.begin(), candidate_uri.size(), candidate) < 0) {
				res = E_FAIL;
			}
		}

		if (res == S_OK && (path == NULL || strcmp(candidate_uri.begin(), path) == 0)) {
			res = captures->BindToObject(IID_IBaseFilter, void_cast(&found));
			break;
		}
	}

	_ASSERT(res == S_OK || res == S_FALSE);
	return res == S_OK ? found : NULL;
}

static long wcstouri(char* dst, long dst_cap, const wchar_t* src) {
	static const char hex[] = "0123456789abcdef";

	long dst_end = 0;

	const wchar_t* c = src;
	if (wcsncmp(c, L"\\\\?\\", 4) == 0) {
		// Discard the extended-length prefix.
		c += 4;
	}

	for (; dst_end + 1 < dst_cap && *c; ++c) {
		if (iswalnum(*c) || wcschr(L"-_.!~*'();?:@&=+$,", *c)) {
			// Neither URI nor UTF-8 requires escaping most of these characters.
			dst[dst_end] = (char) *c;
			dst_end += 1;
		}
		else if (*c < 0x7F) {
			// This character requires a single URI escape.
			if (dst_end + 3 + 1 >= dst_cap) break;

			dst[dst_end+0] = '%';
			dst[dst_end+1] = hex[(*c >> 4) & 0xF];
			dst[dst_end+2] = hex[(*c >> 0) & 0xF];
			dst_end += 3;
		}
		else if (*c < 0x07FF) {
			// UTF-8 encodes this character as two bytes, both of which require escaping.
			if (dst_end + 6 + 1 >= dst_cap) break;

			dst[dst_end+0] = '%';
			dst[dst_end+1] = hex[(*c >> 10) & 0x1 | 0xC];
			dst[dst_end+2] = hex[(*c >> 6) & 0xF];
			dst[dst_end+3] = '%';
			dst[dst_end+4] = hex[(*c >> 4) & 0x3 | 0x8];
			dst[dst_end+5] = hex[(*c >> 0) & 0xF];
		}
		else {
			// TODO
			_ASSERT(FALSE);
		}
	}
	if (dst_end < dst_cap) {
		dst[dst_end] = '\0';
	}
	else {
		_ASSERT(FALSE);
		return -1;
	}
	return dst_end;
}

static int dshow_Close(EVInput* self) {
	return static_cast<DSInput*>(self)->Close();
}

static int dshow_GetFormats(EVInput* self, int start, EVFormat* formats, int formats_cap) {
	return static_cast<DSInput*>(self)->GetFormats(start, formats, formats_cap);
}

static int dshow_SetFormat(EVInput* self, EVFormat* format) {
	return static_cast<DSInput*>(self)->SetFormat(format);
}

static int dshow_GetFormat(EVInput* self, EVFormat* format) {
	return static_cast<DSInput*>(self)->GetFormat(format);
}

static EVInputFrame* dshow_Poll(EVInput* self) {
	return static_cast<DSInput*>(self)->Poll();
}

static EVInputFrame* dshow_Next(EVInput* self) {
	return static_cast<DSInput*>(self)->Next();
}

static int dshow_Callback(EVInput* self, EVICallbackProc callback, void* user) {
	return static_cast<DSInput*>(self)->Callback(callback, user);
}

static const EVFormat* dshow_FrameFormat(EVInputFrame* frame) {
	return static_cast<DSInputFrame*>(frame)->Format();
}

static long dshow_FrameSize(EVInputFrame* frame) {
	return static_cast<DSInputFrame*>(frame)->Size();
}

static void* dshow_FramePointer(EVInputFrame* frame) {
	return static_cast<DSInputFrame*>(frame)->Pointer();
}

static int dshow_FrameDone(EVInputFrame* frame) {
	// Frames are released by their parent, rather than directly.
	return static_cast<DSInput*>(frame->origin)->Done(frame);
}

static int dshow_Geti(EVInput* self, int name, int* value) {
	return static_cast<DSInput*>(self)->Geti(name, value);
}

static int dshow_Getf(EVInput* self, int name, float* value) {
	return static_cast<DSInput*>(self)->Getf(name, value);
}

static int dshow_Seti(EVInput* self, int name, int value) {
	return static_cast<DSInput*>(self)->Seti(name, value);
}

static int dshow_Setf(EVInput* self, int name, float value) {
	return static_cast<DSInput*>(self)->Setf(name, value);
}

static int dshow_GetCapsf(EVInput* self, int name, float* min, float* max, float* step, float* def) {
	return static_cast<DSInput*>(self)->GetCapsf(name, min, max, step, def);
}

static int dshow_GetStopsf(EVInput* self, int name, float* stops, int stops_cap) {
	return static_cast<DSInput*>(self)->GetStopsf(name, stops, stops_cap);
}
