#ifndef EQUIS_VIDEO_FORMAT_H
#define EQUIS_VIDEO_FORMAT_H

#ifndef __cplusplus
extern "C" {
#endif

typedef struct EVFormat EVFormat;

/// Describes the format of an image, frame, or video stream.
struct EVFormat {
    unsigned short types;   ///< Pixel type; one or more EVF_TYPE_* bits.
    unsigned short flags;   ///< Flags; one or more EVF_FLAG_* bits.
    long width;             ///< Width of a frame in pixels.
    long height;            ///< Height of a frame in pixels.
    long usecs;             ///< Average frame delay in microseconds.
};

/// Get the number of bytes in a complete frame.
long evfSize(const EVFormat* format);

/// Get the number of bytes in a single row.
long evfStride(const EVFormat* format);

/// Converts a frame between formats. This function can produce
/// A8R8G8B8, Y8 and Y16 pixel types from any source.
/// If 'dst' and 'src' are equal, will perform the conversion in place.
/// If 'dst' is NULL, will determine whether the conversion is possible.
/// Returns the number of bytes in the destination frame.
long evfConvert(
	EVFormat* dfmt, void* dst, long dcap,
    const EVFormat* sfmt, const void* src, long send);

/// Finds the best match within a list of formats.
/// The 'desired' format will be modified to match that selected.
/// Returns the index of the best format, or -1 if none were compatible.
long evfChoose(
	EVFormat* desired,
	const EVFormat* formats, long end);

/// 32-bpp RGB. Each pixel is encoded as a single 32-bit value,
/// in native byte order. The high 8 bits contain the alpha
/// component, which must be 255 (opaque). The remaining 24 bits
/// contain red, green, and blue. Blue is in the lowest-valued byte.
/// In all cases, 0 = none and 255 = maximum.
#define EVF_TYPE_A8R8G8B8 (0x0001)

/// Unspecified pixel type which includes color information.
#define EVF_TYPE_COLOR (0x0080)

/// Mask including all pixel types with color information.
#define EVF_TYPE_COLOR_MASK (0x00FF)

/// 8-bpp grayscale. Each pixel is a single 8-bit value which
/// encodes a visual brightness. 0 = black, 255 = white.
#define EVF_TYPE_Y8 (0x0100)

// 16-bpp grayscale. Each pixel is a single 16-bit value, in
// native byte order, which encodes a visual brightness.
// 0 = black, 65536 = white.
#define EVF_TYPE_Y16 (0x0200)

/// Unspecified pixel type with grayscale information.
#define EVF_TYPE_GRAY (0x8000)

/// Mask including all pixel types with grayscale information.
#define EVF_TYPE_GRAY_MASK (0xFF00)

/// The first row of the image is the bottom. Otherwise, the first row is the top.
#define EVF_FLAG_BOTTOM_UP (0x0001)

/// Requests the format with the highest resolution.
/// 'width' and 'height' are interpreted as minimum acceptable values.
#define EVF_FLAG_BEST_AREA (0x0002)

/// Requests the format with the highest frame rate.
/// 'usecs' and 'height' is interpreted as a minimum acceptable value.
/// This flag will take precedence if supplied together
/// with EVF_FLAGS_BEST_AREA.
#define EVF_FLAG_BEST_RATE (0x0004)

#ifndef __cplusplus
}
#endif

#endif /* EQUIS_VIDEO_FORMAT_H */
