#ifndef EQUIS_VIDEO_INPUT_H
#define EQUIS_VIDEO_INPUT_H

#include "format.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct EVInput EVInput;
typedef struct EVInputFrame EVInputFrame;

/// Gets the list of available video inputs devices.
/// Each element of the list is NUL-terminated.
/// The list itself is ended with an extra NUL terminator.
/// Returns the number of URIs skipped or written to the list.
int eviEnum(
	int start,
	char* uris, long uris_cap);

long eviGetName(const char* uri,
	char* name, long name_cap);

/// Opens a new video input from the specified URI.
/// If the URI is relative, it will be interpreted as a
/// path starting at the current directory. Otherwise,
/// it must use one of the following schemas:
///
/// dshow:(path)     -- Microsoft Windows DirectShow
/// v4l2:(path)      -- Video for Linux 2
///
EVInput* eviOpen(const char* uri);

/// Opens a local video file using the platform's
/// native path format.
EVInput* eviOpenFile(const char* path);

/// Closes a video input.
int eviClose(EVInput* handle);

/// Gets the formats supported by a video input.
/// Each format will contain a unique combination of
/// flags, width, height, and usecs.
/// Returns the number of formats written to the array.
int eviGetFormats(
	EVInput* handle, int start,
    EVFormat* formats, int formats_cap);

/// Sets the format generated by a video input.
/// Any of the types, width, height, or usecs
/// fields which are zero will be filled with the
/// video input's preferred values.
/// If EVF_FLAG_BEST_AREA is set,
/// the 'width' and/or 'height' may be increased.
/// If EVF_FLAG_BEST_RATE is set,
/// 'usecs' may be decreased.
int eviSetFormat(
	EVInput* handle, EVFormat* format);

/// Gets the actual format generated by a video input.
int eviGetFormat(
	EVInput* handle, EVFormat* format);

/// Gets the next frame from the video input, if it is available.
EVInputFrame* eviPoll(EVInput* handle);

/// Reads the next frame from the video input.
/// This method may block until a frame becomes available,
/// for example on a live capture device or slow disk.
EVInputFrame* eviNext(EVInput* handle);

typedef void (*EVICallbackProc)(EVInputFrame* frame, void* user);

/// Starts receiving callbacks whenever a frame is available.
/// This function may be called again to change or clear the callback.
int eviCallback(EVInput* handle, EVICallbackProc callback, void* user);

EVInputFrame* evifCreateView(const EVFormat* format, void* data, long size);

/// Gets the format of the frame's pixel data.
const EVFormat* evifFormat(EVInputFrame* frame);

/// Gets the number of bytes occupied by frame's pixel data.
long evifSize(EVInputFrame* frame);

/// Gets a pointer to the frame's pixel data.
void* evifPointer(EVInputFrame* frame);

/// Releases a frame. This function must be called for every frame
/// read or received from the callback. No more than two frames may
/// be held by the application.
int evifDone(EVInputFrame* frame);

/// Reads an integer parameter.
int eviGeti(
	EVInput* handle, int name,
    int* value);

/// Reads a floating-point parameter.
int eviGetf(
	EVInput* handle, int name,
	float* value);

/// Writes an integer parameter.
int eviSeti(
	EVInput* handle, int name,
	int value);

/// Writes a floating-point parameter.
int eviSetf(
	EVInput* handle, int name,
	float value);

/// Describes the range of values supported by a floating-point parameter.
///
/// step > 0: Any value y = kx + min where y <= max and k is a non-negative integer.
/// step = 0: All values in the min-max range are acceptable.
/// step < 0: Only specific values are acceptable, call eviGetStopsf.
///
int eviGetCapsf(EVInput* handle, int name,
	float* min, float* max,
    float* step, float* def);

/// Gets the list of values supported by a floating-point parameter.
/// Copies at most 'stops_cap' values into the 'stops' array,
/// and returns the number of values.
/// Returns zero if the parameter is continuous.
///
int eviGetStopsf(EVInput* handle, int name,
	float* stops, int stops_cap);

/// Camera exposure time of each frame. Floating-point seconds.
#define EVI_EXPOSURE (1)

/// Whether the camera automatically adjusts the exposure. Integer boolean.
#define EVI_EXPOSURE_AUTO (101)

/// Camera iris aperture. Floating-point f-stop value.
#define EVI_APERTURE (2)

/// Whether the camera automatically adjusts the aperture. Integer boolean.
#define EVI_APERTURE_AUTO (102)

#ifdef __cplusplus
}
#endif

#endif /* EQUIS_VIDEO_INPUT_H */