#pragma once

#include "Input_Video.hpp"
#include "touch/touch.h"

EQUIS_ENTER_NS3(Equis,Xna,Input);

/// <summary>
/// A press or release detected by the FTIR device.
/// </summary>
/// The mass of an event indicates the size and magnitude of the change. This
/// roughly corresponds to how hard the person is pressing on the table,
/// regardless of the area of effect.
public ref struct FtirEvent
{
	/// <summary>
	/// Creates an event with the specified position and mass. If the mass is
	/// positive, this event is a press. If the mass if native, this event is
	/// a release.
	/// </summary>
	FtirEvent(int type, float x, float y, float mass);

	/// <summary>
	/// Creates an event with the specified position and mass. If the mass is
	/// positive, this event is a press. If the mass if native, this event is
	/// a release.
	/// </summary>
	FtirEvent(int type, Vector2 position, float mass);

	/// <summary>
	/// The calibrated position at which this event occurred.
	/// </summary>
	PROPERTY_RO(Vector2, Position);

	PROPERTY_RO(bool, Press);
	PROPERTY_RO(bool, Release);

	/// <summary>
	/// The calibrated X position at which this event occurred.
	/// </summary>
	PROPERTY_RO(float, X);

	/// <summary>
	/// The calibrated Y position at which this event occurred.
	/// </summary>
	PROPERTY_RO(float, Y);

	/// <summary>
	/// The mass of this event.
	/// </summary>
	PROPERTY_RO(float, Mass);

	PROPERTY_RO(int, SessionId);
	PROPERTY_RO(int, CursorId);
	PROPERTY_RO(int, Frame);
	PROPERTY_RO(int, Since);

	/// <summary>
	/// Creates an event from the unmanaged events used in EquisFTIR-Core.
	/// </summary>
	static FtirEvent ^FromNative(ETEvent *native);

	/// <summary>
	/// Fills an unmanaged event used in EquisFTIR-Core from this event.
	/// </summary>
	void ToNative(ETEvent *native);

private:
	int m_session_id;
	int m_cursor_id;
	Vector2 m_position;
	float m_mass;
	int m_type;
	int m_frame;
	int m_since;
};

/// <summary>
/// Provides access to input data from an FTIR multi-touch sensitive surface.
/// </summary>
public ref class Ftir
{
public:
	/// <summary>
	/// Creates a new FTIR device based on the specified video capture device and format.
	/// </summary>
	/// Textures provided by the preview video streams will be generated from the specified
	/// graphics device manager, and automatically recreated when necessary.
	//static Ftir ^FromCapture(VideoDevice ^captureDevice);

	/// <summary>
	/// Creates a new FTIR device based on the specified video capture device and format.
	/// </summary>
	/// Textures provided by the preview video streams will be generated from the specified
	/// graphics device manager, and automatically recreated when necessary.
	static Ftir ^FromCapture(String ^capturePath);

	/// <summary>
	/// Creates a new FTIR device based on the specified video file.
	/// </summary>
	/// BUG: The video runs far faster than its recorded speed.
	///
	/// Textures provided by the preview video streams will be generated from the specified
	/// graphics device manager, and automatically recreated when necessary.
	static Ftir ^FromFile(String ^filePath);

	/// <summary>
	/// Gets the format in which video data is read from the source camera or file.
	/// </summary>
	/// Previews will be generated with the same resolution and frame rate, but will use
	/// SurfaceFormat.Color (more commonly known as ARGB32).
	PROPERTY_RO(VideoFormat ^, Format);

	PROPERTY_RO(float, RealFramesPerSecond);

#if 0
	/// <summary>
	/// Gets a preview of the input received from the source camera or file.
	/// </summary>
	PROPERTY_RO(Video ^, InputPreview);

	/// <summary>
	/// Gets a preview of the pressure changes interpreted by this device. 
	/// </summary>
	PROPERTY_RO(Video ^, ChangePreview);
#endif

	void PreviewInput(Texture2D^ texture);
	void PreviewFiltered(Texture2D^ texture);

	/// <summary>
	/// Starts processing of the video source.
	/// </summary>
	void Start(void);

	/// <summary>
	/// Stops processing of the video source.
	/// </summary>
	void Stop(void);

	/// <summary> 
	/// Updates the position transform matrix based on a set of calibration points.
	/// </summary>
	/// BUG: The current algorithm (Gaussian elimination) requires exactly four pairs
	///      of points.
	void Calibrate(array<Vector2> ^expectedPositions, array<Vector2> ^actualPositions);

	/// <summary>
	/// The transformation matrix which is applied to all positions generated by this
	/// device.
	/// </summary>
	/// A user interface may choose to distort the preview textures using this matrix,
	/// so that they correspond to the event positions.
	//PROPERTY(Matrix, PositionTransform);

	/// <summary>
	/// Gets the list of events that have occurred since the last time the events were
	/// read.
	/// </summary>
	/// Only events (presses or releases) with at least the specified magnitude of
	/// mass will be returned. The events will be sorted in order of decreasing
	/// magnitude of mass.
	array<FtirEvent ^> ^PollEvents(float requiredMass);

internal:
	Ftir(ETSurface *source);
	~Ftir();

private:
	ETSurface* m_source;
};

EQUIS_LEAVE_NS3(Equis,Xna,Input);
